/******************************************************/
/*  tpax: a topological pax implementation            */
/*  Copyright (C) 2020  Z. Gilboa                     */
/*  Released under GPLv2 and GPLv3; see COPYING.TPAX. */
/******************************************************/

#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>

#define ARGV_DRIVER

#include <tpax/tpax.h>
#include "tpax_version.h"
#include "tpax_driver_impl.h"
#include "argv/argv.h"

/* package info */
static const struct tpax_source_version tpax_src_version = {
	TPAX_TAG_VER_MAJOR,
	TPAX_TAG_VER_MINOR,
	TPAX_TAG_VER_PATCH,
	TPAX_GIT_VERSION
};

/* default fd context */
static const struct tpax_fd_ctx tpax_default_fdctx = {
	.fdin  = STDIN_FILENO,
	.fdout = STDOUT_FILENO,
	.fderr = STDERR_FILENO,
	.fdcwd = AT_FDCWD,
	.fddst = AT_FDCWD,
	.fdlog = (-1),
};

struct tpax_driver_ctx_alloc {
	struct argv_meta *		meta;
	struct tpax_driver_ctx_impl	ctx;
	uint64_t			guard;
	const char *			units[];
};

static uint32_t tpax_argv_flags(uint32_t flags)
{
	uint32_t ret = ARGV_CLONE_VECTOR;

	if (flags & TPAX_DRIVER_VERBOSITY_NONE)
		ret |= ARGV_VERBOSITY_NONE;

	if (flags & TPAX_DRIVER_VERBOSITY_ERRORS)
		ret |= ARGV_VERBOSITY_ERRORS;

	if (flags & TPAX_DRIVER_VERBOSITY_STATUS)
		ret |= ARGV_VERBOSITY_STATUS;

	return ret;
}

static int tpax_driver_usage(
	int				fdout,
	const char *			program,
	const char *			arg,
	const struct argv_option **	optv,
	struct argv_meta *		meta)
{
	char header[512];

	snprintf(header,sizeof(header),
		"Usage: %s [options] <file>...\n" "Options:\n",
		program);

	argv_usage(fdout,header,optv,arg);
	argv_free(meta);

	return TPAX_USAGE;
}

static struct tpax_driver_ctx_impl * tpax_driver_ctx_alloc(
	struct argv_meta *		meta,
	const struct tpax_fd_ctx *	fdctx,
	const struct tpax_common_ctx *	cctx,
	size_t				nunits)
{
	struct tpax_driver_ctx_alloc *	ictx;
	size_t				size;
	struct argv_entry *		entry;
	const char **			units;
	int				elements;

	size =  sizeof(struct tpax_driver_ctx_alloc);
	size += (nunits+1)*sizeof(const char *);

	if (!(ictx = calloc(1,size)))
		return 0;

	memcpy(&ictx->ctx.fdctx,fdctx,sizeof(*fdctx));
	memcpy(&ictx->ctx.cctx,cctx,sizeof(*cctx));

	elements = sizeof(ictx->ctx.erribuf) / sizeof(*ictx->ctx.erribuf);

	ictx->ctx.errinfp  = &ictx->ctx.erriptr[0];
	ictx->ctx.erricap  = &ictx->ctx.erriptr[--elements];

	ictx->meta = meta;

	for (entry=meta->entries,units=ictx->units; entry->fopt || entry->arg; entry++)
		if (!entry->fopt)
			*units++ = entry->arg;

	ictx->ctx.ctx.units = ictx->units;
	ictx->ctx.ctx.errv  = ictx->ctx.errinfp;
	return &ictx->ctx;
}

static int tpax_get_driver_ctx_fail(struct argv_meta * meta)
{
	argv_free(meta);
	return -1;
}

int tpax_get_driver_ctx(
	char **				argv,
	char **				envp,
	uint32_t			flags,
	const struct tpax_fd_ctx *	fdctx,
	struct tpax_driver_ctx ** 	pctx)
{
	struct tpax_driver_ctx_impl *	ctx;
	struct tpax_common_ctx		cctx;
	const struct argv_option *	optv[TPAX_OPTV_ELEMENTS];
	struct argv_meta *		meta;
	struct argv_entry *		entry;
	size_t				nunits;
	const char *			program;

	(void)envp;

	if (!fdctx)
		fdctx = &tpax_default_fdctx;

	argv_optv_init(tpax_default_options,optv);

	if (!(meta = argv_get(
			argv,optv,
			tpax_argv_flags(flags),
			fdctx->fderr)))
		return -1;

	nunits	= 0;
	program = argv_program_name(argv[0]);
	memset(&cctx,0,sizeof(cctx));
	cctx.drvflags = flags;

	if (!argv[1] && (flags & TPAX_DRIVER_VERBOSITY_USAGE))
		return tpax_driver_usage(
			fdctx->fderr,
			program,
			0,optv,meta);

	/* get options, count units */
	for (entry=meta->entries; entry->fopt || entry->arg; entry++) {
		if (entry->fopt) {
			switch (entry->tag) {
				case TAG_HELP:
					if (flags & TPAX_DRIVER_VERBOSITY_USAGE)
						return tpax_driver_usage(
							fdctx->fdout,
							program,entry->arg,
							optv,meta);
					break;

				case TAG_VERSION:
					cctx.drvflags |= TPAX_DRIVER_VERSION;
					break;
			}
		} else
			nunits++;
	}

	if (!(ctx = tpax_driver_ctx_alloc(meta,fdctx,&cctx,nunits)))
		return tpax_get_driver_ctx_fail(meta);

	ctx->ctx.program	= program;
	ctx->ctx.cctx		= &ctx->cctx;

	*pctx = &ctx->ctx;
	return TPAX_OK;
}

static void tpax_free_driver_ctx_impl(struct tpax_driver_ctx_alloc * ictx)
{
	argv_free(ictx->meta);
	free(ictx);
}

void tpax_free_driver_ctx(struct tpax_driver_ctx * ctx)
{
	struct tpax_driver_ctx_alloc *	ictx;
	uintptr_t			addr;

	if (ctx) {
		addr = (uintptr_t)ctx - offsetof(struct tpax_driver_ctx_impl,ctx);
		addr = addr - offsetof(struct tpax_driver_ctx_alloc,ctx);
		ictx = (struct tpax_driver_ctx_alloc *)addr;
		tpax_free_driver_ctx_impl(ictx);
	}
}

const struct tpax_source_version * tpax_source_version(void)
{
	return &tpax_src_version;
}

int tpax_get_driver_fdctx(
	const struct tpax_driver_ctx *	dctx,
	struct tpax_fd_ctx *		fdctx)
{
	struct tpax_driver_ctx_impl *	ictx;

	ictx = tpax_get_driver_ictx(dctx);

	fdctx->fdin  = ictx->fdctx.fdin;
	fdctx->fdout = ictx->fdctx.fdout;
	fdctx->fderr = ictx->fdctx.fderr;
	fdctx->fdlog = ictx->fdctx.fdlog;
	fdctx->fdcwd = ictx->fdctx.fdcwd;
	fdctx->fddst = ictx->fdctx.fddst;

	return 0;
}

int tpax_set_driver_fdctx(
	struct tpax_driver_ctx *	dctx,
	const struct tpax_fd_ctx *	fdctx)
{
	struct tpax_driver_ctx_impl *	ictx;

	ictx = tpax_get_driver_ictx(dctx);

	ictx->fdctx.fdin  = fdctx->fdin;
	ictx->fdctx.fdout = fdctx->fdout;
	ictx->fdctx.fderr = fdctx->fderr;
	ictx->fdctx.fdlog = fdctx->fdlog;
	ictx->fdctx.fdcwd = fdctx->fdcwd;
	ictx->fdctx.fddst = fdctx->fddst;

	return 0;
}
